/*
    Copyright (C) 2001 Hotwarez LLC, Goldtree Enterprises
  
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
    License as published by the Free Software Foundation; 
    version 2 of the License.
  
    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Library General Public License for more details.
  
    You should have received a copy of the GNU Library General Public
    License along with this library; if not, write to the Free
    Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <stdlib.h>
#include "util.h"  /* need exit_gracefully */
#include "keys.h"  /* need Get_Keypress() */
#include "tanks.h"
#include "object.h"
#include "collide.h"
#include <allegro/file.h>

extern char g_DataPath[255];

/* Read a tank, which is a binary dump of struct Vehicle on i386
   from disk, do this in an arch independent way.
   
   Note we use the same brillaint error handling as the original fread
   call: none :) */
void read_tank(Vehicle *tank, PACKFILE *fp)
{
    union {
      int i;
      float f;
    } u;
    float *f;
    int i, *ip;

    /* General information about this vehicle */

    tank->vtype = pack_igetl(fp);
    tank->team = pack_igetl(fp);
    tank->vehicle_mode = pack_igetl(fp);
    tank->alive = pack_igetw(fp);

    /* skip 2 bytes of padding */
    pack_fseek(fp, 2);

    u.i = pack_igetl(fp);
    tank->surface_rad = u.f;

    /* Information about this vehicles 3d object */

    /* skip 4 PointFace pointers */
    pack_fseek(fp, 4 * 4);

    /* BoundingBox contains 6 floats, treat it as an array of floats */
    f = (float *)(&tank->box);
    for (i = 0; i < 6; i++) {
        u.i = pack_igetl(fp);
        f[i] = u.f;
    }

    /* MagicBoundingBox contains 6 32 bit ints, treat it as an array of ints */
    ip = (int *)(&tank->mbox);
    for (i = 0; i < 6; i++) {
        ip[i] = pack_igetl(fp);
    }

    /* skip EdgeTable edge pointer */
    pack_fseek(fp, 4);

    tank->collision_edges.edges = pack_igetl(fp);

    /* Orientation contains 9 floats, treat it as an array */
    f = (float *)(&tank->orient);
    for (i = 0; i < 9; i++) {
        u.i = pack_igetl(fp);
        f[i] = u.f;
    }

    /* Information about this vehicles movement and rotation */

    /* Float_Vector contains 3 floats, treat it as an array */
    f = (float *)(&tank->vel);
    for (i = 0; i < 3; i++) {
        u.i = pack_igetl(fp);
        f[i] = u.f;
    }

    u.i = pack_igetl(fp);
    tank->air_forward_speed = u.f;
    u.i = pack_igetl(fp);
    tank->air_max_forward_speed = u.f;
    u.i = pack_igetl(fp);
    tank->air_inc_forward_speed = u.f;
    u.i = pack_igetl(fp);
    tank->air_max_sidestep_speed = u.f;
    u.i = pack_igetl(fp);
    tank->air_inc_sidestep_speed = u.f;

    u.i = pack_igetl(fp);
    tank->air_rise_rot_speed = u.f;
    u.i = pack_igetl(fp);
    tank->air_spin_rot_speed = u.f;
    u.i = pack_igetl(fp);
    tank->air_inc_rot_speed = u.f;
    u.i = pack_igetl(fp);
    tank->air_max_rot_speed = u.f;

    u.i = pack_igetl(fp);
    tank->surface_max_speed = u.f;
    u.i = pack_igetl(fp);
    tank->surface_inc_speed = u.f;
    u.i = pack_igetl(fp);
    tank->surface_inc_sidestep_speed = u.f;

    u.i = pack_igetl(fp);
    tank->surface_rot_speed = u.f;
    u.i = pack_igetl(fp);
    tank->surface_inc_rot_speed = u.f;
    u.i = pack_igetl(fp);
    tank->surface_max_rot_speed = u.f;

    /* Information about this vehicles weapons */

    /* skip 1 float pointer */
    pack_fseek(fp, 4);

    u.i = pack_igetl(fp);
    tank->laser_speed = u.f;

    tank->laser_life = pack_igetw(fp);
    tank->laser_damage = pack_igetw(fp);
    tank->laser_reload_time = pack_igetw(fp);
    tank->frames_till_fire_laser = pack_igetw(fp);

    u.i = pack_igetl(fp);
    tank->missile_speed = u.f;
    u.i = pack_igetl(fp);
    tank->turning_angle = u.f;

    tank->missile_life = pack_igetw(fp);
    tank->missile_damage = pack_igetw(fp);
    tank->missile_reload_time = pack_igetw(fp);
    tank->frames_till_fire_missile = pack_igetw(fp);
    tank->missile_generation_time = pack_igetw(fp);
    tank->frames_till_new_missile = pack_igetw(fp);
    tank->max_missile_storage = pack_igetw(fp);
    tank->missiles_stored = pack_igetw(fp);

    tank->max_projectiles = pack_igetw(fp);

    /* skip 2 bytes pad + 1 Projectile pointer */
    pack_fseek(fp, 6);

    /* Information about this vehicles hitpoints */

    tank->max_hitpoints = pack_igetl(fp);
    tank->current_hitpoints = pack_igetl(fp);

    tank->ramming_active = pack_igetw(fp);
    tank->ramming_damage = pack_igetw(fp);

    tank->double_lasers_active = pack_igetw(fp);

    tank->mine_reload_time = pack_igetw(fp);
    tank->mine_damage = pack_igetw(fp);
    tank->mine_life = pack_igetw(fp);

    tank->cs_missile_reload_time = pack_igetw(fp);
    tank->cs_missile_life = pack_igetw(fp);

    u.i = pack_igetl(fp);
    tank->cs_missile_speed = u.f;

    tank->controls_scrambled = pack_igetw(fp);
    tank->frames_till_unscramble = pack_igetw(fp);
    tank->scramble_life = pack_igetw(fp);

    tank->traitor_missile_reload_time = pack_igetw(fp);
    tank->traitor_missile_life = pack_igetw(fp);
    /* skip 2 bytes pad */
    pack_fseek(fp, 2);

    u.i = pack_igetl(fp);
    tank->traitor_missile_speed = u.f;

    tank->traitor_life = pack_igetw(fp);
    tank->traitor_active = pack_igetw(fp);
    tank->frames_till_traitor_deactivate = pack_igetw(fp);

    tank->anti_missile_active = pack_igetw(fp);

    tank->cloaking_active = pack_igetw(fp);
    tank->cloak_reload_time = pack_igetw(fp);
    tank->frames_till_cloak = pack_igetw(fp);
    tank->cloak_time = pack_igetw(fp);
    tank->frames_till_cloak_suck = pack_igetw(fp);

    tank->decoy_life = pack_igetw(fp);
    tank->decoy_reload_time = pack_igetw(fp);
    /* skip 2 bytes pad */
    pack_fseek(fp, 2);
}

#define i386_Vehicle_size 316

void Load_Tank( Vehicle *tank, enum VehicleType tank_type )
    {
     PACKFILE *fp;

	char newfilename[512];

	sprintf(newfilename,"%sgamedata/tanks.tbf",g_DataPath);



     if( (fp = pack_fopen( newfilename, "r" )) == NULL ) {
         printf( "Load_Tank() : fopen failed\n" );
         Get_Keypress();
         exit_gracefully();
     }

     pack_fseek( fp, i386_Vehicle_size * tank_type );
     read_tank( tank, fp );

     pack_fclose( fp );

    } /* End of Load_Tank */

void init_vehicle_gradient( Vehicle *tank )
{
    int i;

    if( tank->team == RED_TEAM ) {

        for( i = 0; i < tank->obj->faces; i++ ) {
            if( tank->obj->face[i].gradient == 0 )
                tank->obj->face[i].gradient = RedVehicleFirstGrad;
            else if( tank->obj->face[i].gradient == 1 )
                tank->obj->face[i].gradient = RedVehicleSecondGrad;
            else if( tank->obj->face[i].gradient == 2 )
                tank->obj->face[i].gradient = RedVehicleThirdGrad;
            else
                tank->obj->face[i].gradient = RedVehicleFirstGrad;
        }

        for( i = 0; i < tank->world_obj->faces; i++ ) {
            if( tank->world_obj->face[i].gradient == 0 )
                tank->world_obj->face[i].gradient = RedVehicleFirstGrad;
            else if( tank->world_obj->face[i].gradient == 1 )
                tank->world_obj->face[i].gradient = RedVehicleSecondGrad;
            else if( tank->world_obj->face[i].gradient == 2 )
                tank->world_obj->face[i].gradient = RedVehicleThirdGrad;
            else
                tank->world_obj->face[i].gradient = RedVehicleFirstGrad;
        }

    }
    else {

        for( i = 0; i < tank->obj->faces; i++ ) {
            if( tank->obj->face[i].gradient == 0 )
                tank->obj->face[i].gradient = BlueVehicleFirstGrad;
            else if( tank->obj->face[i].gradient == 1 )
                tank->obj->face[i].gradient = BlueVehicleSecondGrad;
            else if( tank->obj->face[i].gradient == 2 )
                tank->obj->face[i].gradient = BlueVehicleThirdGrad;
            else
                tank->obj->face[i].gradient = BlueVehicleFirstGrad;
        }

        for( i = 0; i < tank->world_obj->faces; i++ ) {
            if( tank->world_obj->face[i].gradient == 0 )
                tank->world_obj->face[i].gradient = BlueVehicleFirstGrad;
            else if( tank->world_obj->face[i].gradient == 1 )
                tank->world_obj->face[i].gradient = BlueVehicleSecondGrad;
            else if( tank->world_obj->face[i].gradient == 2 )
                tank->world_obj->face[i].gradient = BlueVehicleThirdGrad;
            else
                tank->world_obj->face[i].gradient = BlueVehicleFirstGrad;
        }
    }
}



void Init_Tank( Vehicle *tank, enum VehicleType tank_type, team_type team )
    {

    tank->vtype = tank_type;
    tank->team = team;

    if( (tank->obj = (PointFace *) malloc( sizeof( PointFace ) ) ) == NULL ) {
        printf("Init_Tank() : malloc failed\n");
        Get_Keypress();
        exit_gracefully();
    }
    tank->obj->point = NULL;
    tank->obj->points = 0;
    tank->obj->face = NULL;
    tank->obj->faces = 0;

    if( (tank->collision_obj = (PointFace *) malloc( sizeof( PointFace ) ) ) == NULL ) {
        printf("Init_Tank() : malloc failed\n");
        Get_Keypress();
        exit_gracefully();
    }
    tank->collision_obj->point = NULL;
    tank->collision_obj->points = 0;
    tank->collision_obj->face = NULL;
    tank->collision_obj->faces = 0;

    switch( tank_type )
        {
         case Wasp :
             get_object( tank->obj, "3d_data/wasp.pfd" );
             get_object( tank->collision_obj, "3d_data/cwasp.pfd" );
             break;
         case Beetle :
             get_object( tank->obj, "3d_data/beetle.pfd" );
             get_object( tank->collision_obj, "3d_data/cbeetle.pfd" );
             break;
         case Flea :
             get_object( tank->obj, "3d_data/flea.pfd" );
             get_object( tank->collision_obj, "3d_data/cflea.pfd" );
             break;
         case Mosquito :
             get_object( tank->obj, "3d_data/mosquito.pfd" );
             get_object( tank->collision_obj, "3d_data/cmosquit.pfd" );
             break;
         case Spider :
             get_object( tank->obj, "3d_data/spider.pfd" );
             get_object( tank->collision_obj, "3d_data/cspider.pfd" );
             break;
         case Dragonfly :
             get_object( tank->obj, "3d_data/dragon.pfd" );
             get_object( tank->collision_obj, "3d_data/cdragon.pfd" );
             break;
         case Roach :
             get_object( tank->obj, "3d_data/roach.pfd" );
             get_object( tank->collision_obj, "3d_data/croach.pfd" );
             break;
         case Locust :
             get_object( tank->obj, "3d_data/locust.pfd" );
             get_object( tank->collision_obj, "3d_data/clocust.pfd" );
             break;

         default :
             printf("Invalid tank number in Init_Tank \n");
             Get_Keypress();
             exit_gracefully();
             break;

        } /* End switch */

    if( ( tank->world_obj = (PointFace *) malloc( sizeof( PointFace ) ) ) == NULL ) {
        printf("Init_Tank() : malloc failed\n");
        Get_Keypress();
        exit_gracefully();
    }
    tank->world_obj->point = NULL;
    tank->world_obj->points = 0;
    tank->world_obj->face = NULL;
    tank->world_obj->faces = 0;

    init_world_obj( tank );

    init_vehicle_gradient( tank );

    if( ( tank->world_collision_obj = (PointFace *) malloc( sizeof( PointFace ) ) ) == NULL ) {
        printf("Init_Tank() : malloc failed\n");
        Get_Keypress();
        exit_gracefully();
    }
    tank->world_collision_obj->point = NULL;
    tank->world_collision_obj->points = 0;
    tank->world_collision_obj->face = NULL;
    tank->world_collision_obj->faces = 0;

    init_world_collision_obj( tank );

    init_bounding_box( tank );

    tank->collision_edges.edge = NULL;
    tank->collision_edges.edges = 0;

    init_edge_table( tank );

    tank->alive = TRUE;
    tank->target = NULL;
    tank->vehicle_mode = Surface;

} /* End of Init_Tank */

